import type { PermissionName as PermissionNames } from "../Constants";
import { Permissions } from "../Constants";
import Properties from "../util/Properties";
import type { JSONPermission } from "../types/json";

export default class Permission {
    /** The allowed permissions for this permission instance. */
    allow: bigint;
    /** The denied permissions for this permission instance. */
    deny: bigint;
    #json: Record<keyof typeof Permissions, boolean> | undefined;
    constructor(allow: bigint | string, deny: bigint | string = 0n) {
        this.allow = BigInt(allow);
        this.deny = BigInt(deny);
        Properties.looseDefine(this, "#json", undefined, true);
    }

    /** A key-value map of permission to if it's been allowed or denied (not present if neither) */
    get json(): Record<keyof typeof Permissions, boolean> {
        if (!this.#json) {
            const json = {} as Record<keyof typeof Permissions, boolean>;
            for (const perm of Object.keys(Permissions) as Array<keyof typeof Permissions>) {
                if (this.allow & Permissions[perm]) {
                    json[perm] = true;
                } else if (this.deny & Permissions[perm]) {
                    json[perm] = false;
                }
            }

            return (this.#json = json);
        } else {
            return this.#json;
        }
    }

    /**
     * Check if this permissions instance has the given permissions allowed
     * @param permissions The permissions to check for.
     */
    has(...permissions: Array<PermissionNames>): boolean {
        for (const perm of permissions) {
            if (!(this.allow & Permissions[perm])) {
                return false;
            }
        }

        return true;
    }

    toJSON(): JSONPermission {
        return {
            allow: this.allow.toString(),
            deny:  this.deny.toString()
        };
    }

    toString(): string {
        return `[${this.constructor.name} +${this.allow} -${this.deny}]`;
    }
}
